{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Programming Assignment 1: Markov Chains and PageRank"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "collapsed": false
   },
   "outputs": [],
   "source": [
    "%matplotlib inline\n",
    "import matplotlib.pyplot as plt\n",
    "import numpy as np\n",
    "from numpy import linalg as la\n",
    "import networkx as nx\n",
    "print \"Modules Imported!\""
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Recall that a Markov chain is a stochastic signal $X = (X_t)_{t \\in {\\mathbb Z_+}}$, where the distribution of the state $X_{t+1}$ at time $t+1$ is determined by the state $X_t$. The state transitions in a time-homogeneous Markov chain with finite state space $\\{1,2,\\ldots,n\\}$ are fully specified by the $n\\times n$ transition probability matrix $M = (M(x,y))$, where\n",
    "\n",
    "$$\n",
    "M(x,y) = {\\mathbf P}[X_{t+1} = y|X_t = x].\n",
    "$$\n",
    "\n",
    "If the initial state $X_0$ has distribution $p_0$, then the distribution $p_t$ of $X_t$ is obtained by matrix multiplication:\n",
    "\n",
    "$$\n",
    "p_t = p_0 M^t,\n",
    "$$\n",
    "\n",
    "where we represent distributions by row vectors of dimension $n$.\n",
    "\n",
    "The code below creates a Python objects for performing Markov updates of this sort."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "collapsed": true
   },
   "outputs": [],
   "source": [
    "# Markov distribution update object\n",
    "\n",
    "class Markov_dist_update:\n",
    "    def __init__(self, M, p0):   # The transition probability matrix and initial distribution are passed\n",
    "        self.M = M \n",
    "        self.p0 = p0\n",
    "        self.n=np.size(p0)\n",
    "        if np.shape(M) != (self.n,self.n):\n",
    "            print \"Error: Shape of M doesn't match p0\"\n",
    "    \n",
    "    def p(self,t):    # update the distribution of the state at time t\n",
    "        p_new=self.p0\n",
    "        for s in range(t):\n",
    "            p_new = np.dot(p_new,self.M)    # one-step distribution update\n",
    "        return p_new"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "This object has a {\\em state}: it remembers its most recent output. This can be very useful. For example, consider a two-state Markov chain with transition matrix\n",
    "\n",
    "$$\n",
    "M = \\left(\n",
    "\\begin{matrix}\n",
    "0.7 & 0.3 \\\\\n",
    "0.3 & 0.7\n",
    "\\end{matrix}\n",
    "\\right)\n",
    "$$\n",
    "\n",
    "and suppose that the initial state $X_0$ has a ${\\rm Bernoulli}(0.2)$ distribution. The following code uses  `Markov_dist_update` to compute $p_1 = p_0 M$:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "collapsed": false
   },
   "outputs": [],
   "source": [
    "M = np.array([[0.7,0.3],[0.3,0.7]])\n",
    "p0 = np.array([0.8,0.2])\n",
    "next_p = Markov_dist_update(M,p0)\n",
    "print next_p.p(0) # prints the initial distribution passed to Markov_dist_update\n",
    "print next_p.p(1) # prints the one-step update"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Now we are going to compute the two-step update:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "collapsed": false
   },
   "outputs": [],
   "source": [
    "print next_p.p(2)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "... and compare it with the output of $p_2 = p_0 M^2$:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "collapsed": false
   },
   "outputs": [],
   "source": [
    "p2 = np.dot(p0,la.matrix_power(M,2))\n",
    "print p2"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "<b>Exercise 1</b>: in the code cell below, write Python code to compute $p_{15}$ in two ways: using `Markov_dist_update` and the matrix multiplication formula. Compare your results -- they better be the same!"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "collapsed": false
   },
   "outputs": [],
   "source": [
    "########## student code for Exercise 1 here ##########"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Using `matplotlib` and `Markov_dist_update` together, we can visualize the time evolution of $p_t$, as the code below shows:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "collapsed": false
   },
   "outputs": [],
   "source": [
    "# An example of how you can use the class to track the distribution of the state of a Markov chain\n",
    "\n",
    "M=np.array([[0.5,0.5,0,0],[0,0.5,0.5,0],[0,0,0.5,0.5],[0.5,0,0,0.5]])    # set up the Markov matrix\n",
    "p0=np.array([0,0.5,0.5,0])   # specify the initial distribution\n",
    "markov=Markov_dist_update(M,p0)   # create a Markov_dist_update_object\n",
    "\n",
    "nstates = len(p0)   # number of states\n",
    "niter = 30    # number of iterations\n",
    "updates = np.stack((markov.p(s) for s in range(niter)),axis=-1)  # stack all the updates into an array\n",
    "for s in range(nstates):\n",
    "    plt.plot(updates[s])\n",
    "\n",
    "plt.title('Markov chain distribution trajectories')\n",
    "plt.xlabel('Number of iterations')\n",
    "plt.ylabel('Probabilities of the states')"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Feel free to experiment with the above code: change the number of states, Markov transition matrix, initial distribution, etc.\n",
    "\n",
    "_Beware_: `Markov_dist_update` will not throw an error message if $M$ is not a valid Markov matrix or if $p_0$ is not a valid probability distribution!"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Total variation distance and contraction property of Markov updates"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "We will now investigate, by means of simulations, the effect of initial state distribution on the long-term behavior of Markov chains. In particular, given two possible initial distributions $p_0$ and $q_0$ for the initial state $X_0$, we would like to examine how close the iterates $p_t = p_0 M^t$ and $q_t = q_0 M^t$ stay to each other. To that end, let us define the following quantity for any two probability distributions $p$ and $q$ on the same state space ${\\mathsf X}$:\n",
    "\n",
    "$$\n",
    "d_{{\\rm TV}}(p,q) := \\frac{1}{2}\\sum_{x \\in {\\mathsf X}} |p(x) - q(x)|.\n",
    "$$\n",
    "\n",
    "Spend a few minutes to convince yourselves that it obeys the following:\n",
    "\n",
    "<ol>\n",
    "<li> $d_{{\\rm TV}}(p,q) \\ge 0$ for all $p,q$\n",
    "<li> $d_{{\\rm TV}}(p,q) = 0$ if and only if $p \\equiv q$\n",
    "<li> $d_{{\\rm TV}}(p,q) \\le d_{{\\rm TV}}(p,r) + d_{{\\rm TV}}(r,q)$ for any three probability distributions $p,q,r$\n",
    "</ol>\n",
    "\n",
    "In other words, $d_{{\\rm TV}}$ can be thought of as a _distance_ between probability distributions: it is nonnegative; equals zero if and only if the two distributions are exactly the same (i.e., assign the same probability to each state); and it satisfies the triangle inequality. For this reason, it has a name -- _total variation distance_."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "<b>Problem 1:</b> In the markdown cell below, prove that $d_{{\\rm TV}}(p,q)$ always takes values between $0$ and $1$, and find the necessary and sufficient condition for $d_{{\\rm TV}}(p,q) = 1$."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "--- student answer for Problem 1 goes here ---"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "<b>Problem 2:</b> Consider a four-state Markov chain with transition matrix\n",
    "\n",
    "$$\n",
    "M = \\left(\n",
    "\\begin{matrix}\n",
    "2/3 & 1/3 & 0 0 \\\\\n",
    "0 & 2/3 & 1/3 & 0 \\\\\n",
    "0 & 0 & 2/3 & 1/3 \\\\\n",
    "1/3 & 0 & 0 & 2/3\n",
    "\\end{matrix}\n",
    "\\right).\n",
    "$$\n",
    "\n",
    "Pick any two initial state distributions $p_0$ and $q_0$ with $d_{{\\rm TV}}(p_0,q_0) = 1$. In the cell below, write Python code based on `Markov_dist_update` to plot the total variation distance $d_{{\\rm TV}}(p_t,q_t)$ for $t = 0, 1,\\ldots, 30$. "
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "collapsed": true
   },
   "outputs": [],
   "source": [
    "######## student code for Problem 2 goes here #########"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "What do you observe? Does $d_{\\rm TV}(p_t,q_t)$ increase or decrease with time? Experiment with changing $M$, $p_0$, or $q_0$ -- does this behavior persist? What if you choose $p_0$ and $q_0$ with $0 < d_{\\rm TV}(p_0,q_0) < 1$?"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "--- student answer here ---"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "In the markdown cell below, prove the _contraction property of Markov chains_: for any two initial probability distributions $p_0$ and $q_0$,\n",
    "\n",
    "$$\n",
    "d_{\\rm TV}(p_{t+1},q_{t+1}) \\le d_{\\rm TV}(p_t,q_t).\n",
    "$$"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "--- student answer here ---"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## PageRank ##"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "The PageRank algorithm discussed in class finds a vector of webpage ranks satisfying a certain self-consistency condition based on the hyperlink graph. In this part of the assignment, you will investigate the convergence properties of PageRank. The main idea behind PageRank is that the desired ranking is an invariant probability distribution of a certain Markov chain. Thus, we consider the state space ${\\mathsf X} = \\{1,\\ldots,n\\}$, where $n$ is the number of webpages of interest. We start by building the _web graph_: each webpage $i \\in {\\mathsf X}$ is associated with a vertex, and we draw an arrow (also called a directed edge or an arc) from vertex $i$ to vertex $j$ if and only if page $i$ contains a hyperlink to page $j$. The cell below uses the `networkx` module to generate a random web graph.\n",
    "\n",
    "Specifically, we use the simplest model of a random graph, the so-called _Erd&#x0151;s--R&eacute;nyi graph_. Such a graph has $n$ vertices, and the edges between vertices are added at random. In the undirected version, denoted by $G(n,p)$, an edge between any two distinct vertices $i$ and $j$ is included with probability $p$ independently of all other edges. In the directed version, denoted by $G_d(n,p)$, the edges from $i$ to $j$ and from $j$ to $i$ are treated separately, and each is included with probability $p$, again independently of all others. The code below generates and displays a $G_d(10,0.5)$ random graph. You should experiment with it by changing the number of vertices and the edge inclusion probability."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "collapsed": false
   },
   "outputs": [],
   "source": [
    "n = 10\n",
    "p = 0.5\n",
    "network1 = nx.erdos_renyi_graph(n,p, seed=None, directed=True)\n",
    "plt.figure(figsize=(10,10))\n",
    "nx.draw(network1)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "The above code is slow when $n$ is large and $p$ is close to $1$. The code below is optimized to generate sparse random graphs, where $p$ is small."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "collapsed": false
   },
   "outputs": [],
   "source": [
    "n = 100\n",
    "p = 0.2\n",
    "network2 = nx.fast_gnp_random_graph(n,p, seed=None, directed=True)\n",
    "plt.figure(figsize=(10,10))\n",
    "nx.draw(network2)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "The code below generates a random webgraph and constructs the Markov matrix corresponding to a Random Surfer Model on that graph.\n",
    "\n",
    "Once the random webgraph is generated, we extract its _adjacency matrix_ $A$. This is an $n\\times n$ matrix, where the entry in row $i$ and column $j$ is $1$ if there is an edge from $i$ to $j$, and $0$ otherwise. The Python function `nx.adjacency_matrix()` takes a graph object as input and returns its adjacency matrix in the so-called _sparse matrix_ format, i.e., as a list of (row,column) pairs where the matrix has a nonzero entry. We use `todense()` function to convert the sparse matrix to the usual `numpy` array. Then we create the Markov matrix $S$ corresponding to the Random Surfer Model on our random webgraph. The number of vertices $n$ is deliberately set to a small value, so that the matrix $S$ can be easily displayed. You should experiment with this code by changing `num_pages` and `link_prob`.\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "collapsed": false
   },
   "outputs": [],
   "source": [
    "num_pages = 10   # specify the number of webpages\n",
    "link_prob = 0.2  # specify the probability of an outgoing link\n",
    "webgraph = nx.fast_gnp_random_graph(num_pages, link_prob, seed=None, directed=True)   # generate a random webgraph\n",
    "A = nx.adjacency_matrix(webgraph)    # extract the adjacency matrix\n",
    "B = np.array(A.todense())            # convert to dense np.array\n",
    "row_sums = B.sum(axis=1)             # compute row sums\n",
    "a = np.zeros(num_pages)              \n",
    "a[row_sums == 0] = 1                 # create a vector that keeps track of pages with no outgoing links\n",
    "\n",
    "#### now we generate the Random Surfer Model matrix ####\n",
    "E = np.zeros((num_pages,num_pages))\n",
    "E[:,0] = a                          \n",
    "V = np.add(B,E)\n",
    "row_sums = V.sum(axis=1)\n",
    "S = V.astype(np.float) / row_sums.reshape(num_pages,1)\n",
    "print S"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "We are now ready to implement a first attempt at PageRank. The code below generates a large, sparse random webgraph and plots the evolution of the PageRank iterates. We start by ranking all the pages equally (i.e., $p_0$ is the uniform distribution on $\\{1,\\ldots,n\\}$) and generate the PageRank iterates $p_t = p_0 S^t$. You should run this code multiple times to convince yourself that PageRank either produces oscillating iterates or converges to the trivial ranking where all webpages except one are assigned zero rank."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "collapsed": false
   },
   "outputs": [],
   "source": [
    "num_pages = 200    # specify the number of webpages\n",
    "link_prob = 0.001  # specify the probability of an outgoing link\n",
    "webgraph = nx.fast_gnp_random_graph(num_pages, link_prob, seed=None, directed=True)   # generate a random webgraph\n",
    "A = nx.adjacency_matrix(webgraph)    # extract the adjacency matrix\n",
    "B = np.array(A.todense())            # convert to dense np.array\n",
    "row_sums = B.sum(axis=1)             # compute row sums\n",
    "a = np.zeros(num_pages)              \n",
    "a[row_sums == 0] = 1                 # create a vector that keeps track of pages with no outgoing links\n",
    "\n",
    "#### now we generate the Random Surfer Model matrix ####\n",
    "E = np.zeros((num_pages,num_pages))\n",
    "E[:,0] = a                          \n",
    "V = np.add(B,E)\n",
    "row_sums = V.sum(axis=1)\n",
    "S = V.astype(np.float) / row_sums.reshape(num_pages,1)\n",
    "\n",
    "#### first attempt at PageRank #####\n",
    "p0 = np.ones(num_pages) / num_pages     # initialize with the uniform distribution\n",
    "PR = Markov_dist_update(S,p0)           # instantiate a Markov_dist_update object\n",
    "nstates = len(p0)   # number of states\n",
    "niter = 30    # number of iterations\n",
    "updates = np.stack((PR.p(s) for s in range(niter)),axis=-1)  # stack all the updates into an array\n",
    "for s in range(nstates):\n",
    "    plt.plot(updates[s])\n",
    "plt.title('PageRank trajectory')\n",
    "plt.xlabel('Number of iterations')\n",
    "plt.ylabel('Probabilities of the states')"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "<b>Problem 3:</b> In the code cell below, modify the above code to construct the Google matrix $G$, as described in the lecture notes. Use the setting $\\alpha = 0.15$. You should now see that the PageRank algorithm always converges, and that the convergence is rather rapid. Again, run the code several times to convince yourself that this is indeed the case."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "collapsed": true
   },
   "outputs": [],
   "source": [
    "##### student code for Problem 3 here #####"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "<b>Problem 4:</b> Now we will investigate the effect of replacing the Random Surfer Model matrix $S$ with the Google matrix $G$. In the markdown cell below, prove that, for any two probability distributions $p$ and $q$ on the state space ${\\mathsf X} = \\{1,\\ldots,n\\}$, we have the _strict contraction_\n",
    "\n",
    "$$\n",
    "d_{{\\rm TV}}(p G, q G) \\le (1-\\alpha) d_{\\rm TV}(p,q).\n",
    "$$"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "--- student answer here ---"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Now consider running PageRank with an arbitrary initial ranking $p_0$. In the markdown cell below, prove that, for any $t$,\n",
    "\n",
    "$$\n",
    "d_{\\rm TV}(p_t, p_t G) \\le (1-\\alpha)^t.\n",
    "$$\n",
    "\n",
    "What does this say about the speed of convergence toward the desired ranking?\n",
    "\n",
    "_Hint:_ use the fact that $p_t = p_0 G^t$ and $p_t G = p_0 G^{t+1} = (p_0 G) G^t$ and the strict contraction property."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "--- student answer here ---"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "<b>Problem 5:</b> In the code cell below, modify your code from Problem 4 to plot the total variation distance $d_{\\rm TV}(p_t,p_t G)$ versus time. Do you observe the rapid convergence?"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "collapsed": true
   },
   "outputs": [],
   "source": [
    "###### student code for Problem 5 here #######"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python [conda root]",
   "language": "python",
   "name": "conda-root-py"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 2
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython2",
   "version": "2.7.12"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 1
}
